/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <assert.h>

#include "detect.h"

#define BIT(x) (1 << (x))

#define SET_PORT(port_tbl, port) \
        ( (port_tbl[(port - 1) / 8]) |= BIT(((port - 1) % 8)))

#define IS_PORT_SET(port_tbl, port) \
        ( (port_tbl[(port - 1) / 8]) & BIT(((port - 1) % 8)))


static plugin_detect_t plugin;
static subscribtion_t subscribtion[] = {
        { p_ip, NULL  },
        { p_tcp, NULL },
        { p_udp, NULL },
        { p_end, NULL }
};


typedef struct _cnxInfo 
{
        Packet_t *packet;
        
        unsigned char port[65536 / 8];
        unsigned int cnx_count;
        unsigned int first_port;
        unsigned int last_port;

        char *kind;
        long first_cnxtime;
        long last_cnxtime;
        timer_t timer;
} cnxInfo_t;


static unsigned int cnx_ttl = 0;
static unsigned int max_cnx_count = 0;
static unsigned int plug_id;
static unsigned int dportcur;
static const struct ip *ipcur;



/*
 *
 */
static void do_report_if_needed(cnxInfo_t *cnx) 
{
        if ( cnx->cnx_count >= max_cnx_count ) {
                struct timeval tv;
                
                gettimeofday(&tv, NULL);
                
                plugin_do_report((plugin_generic_t *) &plugin,
                                 cnx->packet, normal, "Scanning attack",
                                 "%s scanning attempt: %u cnx from port "
                                 "%u to %u in %ld seconds.",
                                 cnx->kind, cnx->cnx_count,
                                 cnx->first_port, cnx->last_port,
                                 tv.tv_sec - cnx->first_cnxtime);
        }    
}



/*
 *
 */
static void expire_cnx(unsigned long arg) 
{
        hostdb_t *h = (hostdb_t *) arg;
        cnxInfo_t *cnx = (cnxInfo_t *) hostdb_get_plugin_data(h, plug_id);
        
        timer_destroy(&cnx->timer);
        hostdb_del(h, plug_id);

        do_report_if_needed(cnx);

        packet_release(cnx->packet);
        free(cnx->kind);
        free(cnx);
}



/*
 *
 */
static const char *kind_cnxInfo(unsigned char flags) 
{
        if ( flags & TH_ACK )
                switch(flags) {
                        
                case (TH_SYN | TH_RST | TH_ACK | TH_FIN | TH_PUSH | TH_URG):
                        return "Xmas";
                default:
                        return NULL;

                }
        
        
        switch(flags) {
                
        case TH_RST:
                break;
                
        case TH_SYN :
                return "Syn";
                
        case TH_FIN:
                return "Fin";
                
        case 0:
                return "Null";
                                               
        case (TH_FIN | TH_PUSH | TH_URG):
                return "Xmas";
                
        default:
                return "Unknow";

        }

        return NULL;
}



/*
 *
 */
static cnxInfo_t *new_cnx(Packet_t *packet, const char *kind, hostdb_t *h) 
{
        cnxInfo_t *tmp;
        struct timeval tv;
        
        tmp = (cnxInfo_t *) malloc(sizeof(cnxInfo_t));
        if ( ! tmp ) {
                log(LOG_ERR, "malloc");
                return NULL;
        }

        packet_lock(packet);
        gettimeofday(&tv, NULL);
        memset(tmp->port, 0, sizeof(tmp->port));
        
        tmp->port[dportcur / 8] |= BIT(dportcur % 8);
        tmp->last_cnxtime = tmp->first_cnxtime = tv.tv_sec;
        tmp->last_port = tmp->first_port = dportcur;
        tmp->cnx_count = 1;
        tmp->packet = packet;
        tmp->kind = strdup(kind);
                
        timer_set_expire(&tmp->timer, cnx_ttl);
        timer_set_data(&tmp->timer, h);
        timer_set_callback(&tmp->timer, expire_cnx);
        
        timer_init(&tmp->timer);

        return tmp;
}



/*
 *
 */
static void modify_cnx(cnxInfo_t *cnx) 
{
        int ret;
        unsigned int p = dportcur - 1;
        struct timeval tv;
        
        gettimeofday(&tv, NULL);

        cnx->last_cnxtime = tv.tv_sec;

        if ( dportcur < cnx->first_port )
                cnx->first_port = dportcur;
        else if ( dportcur > cnx->last_port )
                cnx->last_port = dportcur;
        
        ret = cnx->port[p / 8] & BIT(p % 8);
        if ( ret == 0 ) {
                timer_reset(&cnx->timer);
                cnx->port[p / 8] |= BIT(p % 8);
                cnx->cnx_count++;
        }
}



/*
 *
 */
static void update_hdb_entry(Packet_t *packet, const char *kind, hostdb_t *h) 
{
        cnxInfo_t *tmp;

        /*
         * If this plugin didn't registered any data in the host
         * database at this time, create a new cnxInfo_t structure.
         */
        tmp = (cnxInfo_t *) hostdb_get_plugin_data(h, plug_id);
        if ( ! tmp ) {
                tmp = new_cnx(packet, kind, h);
                if ( ! tmp )
                        return;
                
                hostdb_set_plugin_data(h, plug_id, (unsigned long)tmp);
        } else 
                modify_cnx(tmp); 
}



/*
 *
 */
static void create_hdb_entry(Packet_t *packet, const char *kind) 
{
        hostdb_t *hdb;
        cnxInfo_t *tmp;
        
        hdb = hostdb_new(ipcur);
        if (! hdb)
                return;

        tmp = new_cnx(packet, kind, hdb);
        if ( ! tmp )
                return;
        
        hostdb_set_plugin_data(hdb, plug_id, (unsigned long)tmp);
}



/*
 *
 */
static void generic_packet(Packet_t *packet, const char *kind)
{
        hostdb_t *hdb;
        
        if ( dportcur <= 0 || dportcur > 65535 )
                return;

        hdb = hostdb_search(ipcur);
        if ( ! hdb ) 
                create_hdb_entry(packet, kind);
        else 
                update_hdb_entry(packet, kind, hdb);
}



/*
 *
 */
static void tcp_packet(Packet_t *packet, const struct tcphdr *tcp)
{
        const char *kind;
        
        kind = kind_cnxInfo(tcp->th_flags);
        if ( ! kind )
                return;

        dportcur = ntohs(tcp->th_dport);
        generic_packet(packet, kind);
}



/*
 *
 */
static void udp_packet(Packet_t *packet, const struct udphdr *udp)
{
        dportcur = ntohs(udp->uh_dport);
        generic_packet(packet, "Udp");
}



/*
 *
 */
static void scandetect_run(Packet_t *packet, int depth) 
{
        const struct tcphdr *tcp;
        const struct udphdr *udp;
        
        switch (packet[depth].proto) {
        case p_ip:
                ipcur = &packet[depth].p.ip->ip_hdr;
                break;

        case p_tcp:
                tcp = &packet[depth].p.tcp->tcp_hdr;
                tcp_packet(packet, tcp);
                break;

        case p_udp:
                udp = packet[depth].p.udp_hdr;
                udp_packet(packet, udp);
                break;

        default:
                log(LOG_ERR, "Unknow protocol %d", packet[depth].proto);
                break;
        }
}



static void print_help(void) 
{
        fprintf(stderr, "Usage for %s :\n", plugin.name);
        fprintf(stderr, "\t -c --cnx-count (default=%d) "
                "Number of cnx to track before issuing report\n", max_cnx_count);
        fprintf(stderr, "\t -t --cnx-ttl (default=%d) "
                "Obsolete a cnx if no packet from this cnx are received in X seconds.\n\n",
                cnx_ttl);
}



static int check_opts(void) 
{
        int c, argc;
        char **argv;
        config_t *cfg;
        const char *str;
        struct option opts[] = {
                { "cnx-count", required_argument, NULL, 'c' },
                { "cnx-ttl", required_argument, NULL, 't' },
                { "help", no_argument, NULL, 'h' },
                { 0, 0, 0, 0 },
        };

        plugin_get_opts(plugin.name, &argc, &argv);
        
        while ( (c = getopt_long(argc, argv, "c:t:h", opts, NULL) ) != -1 ) {
                switch (c) {
                case 'c':
                        max_cnx_count = atoi(optarg);
                        break;

                case 't':                        
                        cnx_ttl = atoi(optarg);
                        break;
                        
                case 'h':
                        print_help();
                        return -1;
                        
                default:
                        return -1;
                }
        }
        
        if ( max_cnx_count && cnx_ttl )
                return 0;

        cfg = config_open(PRELUDE_CONF);
        if ( ! cfg ) 
                log(LOG_ERR, "couldn't open %s.\n", PRELUDE_CONF);
        
        if ( ! max_cnx_count ) {
                if ( cfg && (str = config_get(cfg, plugin.name, "max_cnx_count")) )
                        max_cnx_count = atoi(str);
                else
                        max_cnx_count = 30;
        }
        
        if ( ! cnx_ttl ) {
                if ( cfg && (str = config_get(cfg, plugin.name, "cnx_ttl")) )
                        cnx_ttl = atoi(str);
                else
                        cnx_ttl = 5;
        }

        if ( cfg )
                config_close(cfg);
        
        return 0;
}


int plugin_init(unsigned int id)
{        
        plug_id = id;

        plugin_set_name(&plugin, "ScanDetect");
        if ( check_opts() < 0 )
                return -1;
        
        plugin_set_author(&plugin, "Yoann Vandoorselaere");
        plugin_set_contact(&plugin, "yoann@mandrakesoft.com");
        plugin_set_desc(&plugin, "Detect almost all kind of scanning.");
        plugin_set_running_func(&plugin, scandetect_run);
        plugin_set_subscribtion(&plugin, subscribtion);
        
        return plugin_register((plugin_generic_t *)&plugin);
}



