/*****
*
* Copyright (C) 1998 - 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>

#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "alert-common.h"
#include "common.h"
#include "auth.h"
#include "cnx.h"
#include "report-infos.h"
#include "plugin-common.h"
#include "plugin-report.h"
#include "pconfig.h"
#include "ssl.h"
#include "socket-op.h"

#ifdef HAVE_XDR

#include <rpc/types.h>
#include <rpc/xdr.h>
#include "rxdr.h"

static XDR xdrs;

#endif

static ssize_t (*my_read)(int fd, void *buf, size_t count);
static Packet_t packet[MAX_PKTDEPTH];
extern struct list_head __report_plugins;
extern struct report_config config;



#ifdef HAVE_XDR
static int is_endian_convertion_needed(void) 
{
        return config.use_xdr;
}
#endif

/*
 *
 */
static int wait_raw_report(int socket) 
{
        int ret;
        alert_t alert;
        report_infos_t rinfos;

        alert_packet(&alert) = packet;
        
        while ( 1 ) {            
                ret = alert_read(socket, &alert, my_read);
                if ( ret <= 0 )
                        return ret;
                
                report_infos_get(&alert, &rinfos);
                report_plugins_run(&alert, &rinfos);
                report_infos_free(&rinfos);
                
                alert_free(&alert, 1);

        }

        return 0;
}



/*
 *
 */
#ifdef HAVE_XDR

static int wait_xdr_report(int sock) 
{
        int ret;
        alert_t alert;
        report_infos_t rinfos;

        while ( 1 ) {
                if ( ! xdrrec_skiprecord(&xdrs) ) {
                        log(LOG_ERR, "There was an error waiting / reading next report.\n");
                        return -1;
                }

                memset(&alert, 0, sizeof(alert_t));
                memset(packet, 0, sizeof(packet));
                alert_packet(&alert) = packet;
                
                ret = rxdr_decode(&alert, &xdrs);
                if ( ret < 0 )
                        return -1;
                
                report_infos_get(&alert, &rinfos);
                report_plugins_run(&alert, &rinfos);
                report_infos_free(&rinfos);
                
                alert_free(&alert, 0);

        }

        return 0;  
}



static int read_xdr_report(char *handle, char *buf, int len) 
{
        int ret;
        int sock = *((int *)handle);

        ret = my_read(sock, buf, len);
        return (ret == 0) ? -1 : ret;
}

#endif




static int set_options(const char *optbuf) 
{
        if ( strstr(optbuf, "use_ssl=yes;") ) {
#ifndef HAVE_SSL
                goto unavaillable;
#else
                log(LOG_INFO, "\t- Client requested SSL communication.\n");
                config.use_ssl = 1;
#endif          
        } else {
#ifdef HAVE_SSL
                log(LOG_INFO, "\t- Client requested non encrypted communication.\n");
#endif
                config.use_ssl = 0;
        }


        if ( strstr(optbuf, "use_xdr=yes;") ) {
#ifndef HAVE_XDR
                goto unavaillable;
#else
                log(LOG_INFO, "\t- Client requested XDR to be used.\n");
#endif
                config.use_xdr = 1;
        }
        
        return 0;
        
#if ! defined(HAVE_XDR) || ! defined(HAVE_SSL)
 unavaillable:
        log(LOG_INFO, "\t- Client requested unavaillable option.\n");
        return -1;
#endif
}




static int setup_connection(int sock) 
{
        int ret, len;
        char buf[1024];
        const char *ssl, *xdr;
        
#ifdef HAVE_SSL
        ssl = "supported";
#else
        ssl = "unsupported";
#endif

#ifdef HAVE_XDR
        xdr = "supported";
#else
        xdr = "unsupported";
#endif

        len = snprintf(buf, sizeof(buf), "xdr=%s; ssl=%s;\n", xdr, ssl);

        ret = socket_write_delimited(sock, buf, ++len);
        if ( ret < 0 ) {
                log(LOG_ERR, "error writing config to Prelude client.\n");
                return -1;
        }
        
        ret = socket_read_delimited(sock, buf, sizeof(buf));
        if ( ret < 0 ) {
                log(LOG_ERR, "error reading Prelude client config string.\n");
                return -1;
        }


        return set_options(buf);
}




/*
 *
 */
int handle_inet_connection(int sock, struct sockaddr *addr, unsigned int *addrlen) 
{
        int ret;
        char *from;

	from = strdup(inet_ntoa(((struct sockaddr_in *)addr)->sin_addr));
	if ( ! from )
		return -1;

        log(LOG_INFO, "new connection from %s.\n", from);

        ret = setup_connection(sock);
        if ( ret < 0 )
                goto err;
        
        my_read = read;
        
#ifdef HAVE_SSL
        if ( config.use_ssl == 1 ) {
                my_read = ssl_read;
                
                ret = ssl_auth_client(sock);
                if ( ret < 0 ) {
                        log(LOG_INFO, "SSL authentication failed with %s.\n", from);
                        goto err;
                }
                
                log(LOG_INFO, "SSL authentication suceeded with %s.\n", from);
        } else
                
#endif
                if ( auth_check(sock) < 0 ) 
                        goto err;
                
#ifdef HAVE_XDR
        ret = is_endian_convertion_needed();
        if ( ret ) {
                xdrs.x_op = XDR_DECODE; 
                xdrrec_create(&xdrs, 0, 0, (char *)&sock, read_xdr_report, NULL);
                ret = wait_xdr_report(sock);
                xdr_destroy(&xdrs);
        } else
#endif
                ret = wait_raw_report(sock);

        log(LOG_INFO, "closing connection with %s.\n", from);
        free(from);

        return ret;

  err:
         log(LOG_INFO, "closing connection with %s.\n", from);
         free(from);
         close(sock);
         return -1;
}


/*
 *
 */
int handle_unix_connection(int sock, struct sockaddr *addr, unsigned int *addrlen) 
{
        int ret;
        
        log(LOG_INFO, "new local connection.\n");

        my_read = read;
        ret = wait_raw_report(sock);

        log(LOG_INFO, "closing local connection.\n");

        return ret;
}














