/*****
*
* Copyright (C) 2001 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <pthread.h>

#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "common.h"
#include "async-write.h"
#include "pconfig.h"
#include "rwrite.h"
#include "detect-plugins-api.h"


#define MAX_IO 800


static LIST_HEAD(aio_queue);
static int aio_count = 0;
static pthread_t aio_thr;
static pthread_cond_t input_possible_cond, output_possible_cond;
static pthread_mutex_t input_possible_mutex, output_possible_mutex;


static void remove_aio_item(alert_t *item) 
{        
        list_del(&item->list);
        
        pthread_mutex_lock(&input_possible_mutex);
        if ( aio_count-- == MAX_IO ) 
                pthread_cond_signal(&input_possible_cond);
        pthread_mutex_unlock(&input_possible_mutex);
}




static void add_aio_item(alert_t *alert) 
{
        list_add_tail(&alert->list, &aio_queue);
        
        pthread_mutex_lock(&output_possible_mutex);
        if ( aio_count++ == 0 ) 
                pthread_cond_signal(&output_possible_cond);
        pthread_mutex_unlock(&output_possible_mutex);
}



static void flush_aio_queue(void) 
{
        alert_t *alert;
        struct list_head *tmp;

        for ( tmp = aio_queue.next; tmp != &aio_queue; ) {
                alert = list_entry(tmp, alert_t, list);

                tmp = tmp->next;
                remove_aio_item(alert);

                rwrite_write(alert, 1);
                packet_release(alert_packet(alert));
                free(alert);
        }
}




static void *aio_thread(void *null) 
{
        int ret;
        sigset_t set;
        
        sigemptyset(&set);
        sigaddset(&set, SIGPIPE);
        
        ret = pthread_sigmask(SIG_BLOCK, &set, NULL);
        if ( ret < 0 ) {
                log(LOG_ERR, "pthread_sigmask returned an error.\n");
                return NULL;
        }
        
        while ( 1 ) {
                pthread_mutex_lock(&output_possible_mutex);
                while ( aio_count == 0 ) 
                        pthread_cond_wait(&output_possible_cond, &output_possible_mutex);
                pthread_mutex_unlock(&output_possible_mutex);

                flush_aio_queue();
        }
}



/**
 * async_write:
 * @alert: the alert to send.
 *
 * This queue the @alert to be written ASAP using
 * the rwrite_write() function.
 */
void async_write(alert_t *alert) 
{
        pthread_mutex_lock(&input_possible_mutex);
        while ( aio_count == MAX_IO ) 
                pthread_cond_wait(&input_possible_cond, &input_possible_mutex);
        pthread_mutex_unlock(&input_possible_mutex);

        add_aio_item(alert);
}



/**
 * async_init:
 *
 * This function initialize the asynchronous IO thread.
 *
 * Returns: 0 on success, -1 on error.
 */
int async_init() 
{
        int ret;

        log(LOG_INFO, "Async I/O Reserved %u bytes (%u elements * %u bytes) of memory.\n",
            MAX_IO * sizeof(alert_t), MAX_IO, sizeof(alert_t));
        
        pthread_cond_init(&input_possible_cond, NULL);
        pthread_cond_init(&output_possible_cond, NULL);
        pthread_mutex_init(&input_possible_mutex, NULL);
        pthread_mutex_init(&output_possible_mutex, NULL);
                
        ret = pthread_create(&aio_thr, NULL, aio_thread, NULL);
        if ( ret < 0 ) {
                log(LOG_ERR, "couldn't create asynchronous IO thread : %m.\n");
                return -1;
        }
          
        return 0;
}



/**
 * async_exit:
 *
 * Kill the asynchronous IO thread,
 * and flush all pending IO.
 */
void async_exit(void) 
{
        pthread_cancel(aio_thr);
        flush_aio_queue();
        
        pthread_cond_destroy(&input_possible_cond);
        pthread_cond_destroy(&output_possible_cond);
        
        pthread_mutex_destroy(&input_possible_mutex);
        pthread_mutex_destroy(&output_possible_mutex);
}






