/*****
*
* Copyright (C) 1998 - 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/poll.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <pcap.h>


#include "pconfig.h"
#include "capture.h"
#include "packet-decode.h"
#include "timer.h"
#include "common.h"
#include "packet.h"


typedef struct {
        char *bpf;
        pcap_t *pd;
        int pcap_filedes;
        bpf_u_int32 netmask;
        pcap_handler p_handler;
} capdev_t;


static unsigned long long packet_counter = 0;
static int interfaces_nbr = 0;
static capdev_t **devices = NULL, *file_dev = NULL;
static struct pollfd *pfd;
static const char *capture_to = NULL;

extern unsigned long prelude_reporting_time;
extern Pconfig_t config;


/*
 * Search for the datalink handler corresponding
 * to the datalink of type 'type'.
 *
 * Return NULL if no handler was found.
 */
static pcap_handler search_datalink_handler(int type)
{
        int i;
        struct datalink_handler {
                pcap_handler handler;
                int type;
        } p[] = {
                {	SliceAndStoreEtherPkt, DLT_EN10MB	},
                {	SliceAndStoreEtherPkt, DLT_IEEE802	},
                {       SliceAndStoreNullPkt,  DLT_NULL         },
                {       SliceAndStoreRawPkt,   DLT_RAW          },
#ifdef DLT_LOOP
                {       SliceAndStoreNullPkt,  DLT_LOOP         },
#endif
                {       SliceAndStorePppPkt,   DLT_PPP          },
                {       SliceAndStorePppBsdosPkt, DLT_PPP_BSDOS },
                {       SliceAndStoreFddiPkt, DLT_FDDI          },
                {       SliceAndStoreAtmPkt, DLT_ATM_RFC1483    },
                {       NULL, 0                                 }, 
        };

        for (i = 0; p[i].handler; i++ ) 
                if ( type == p[i].type ) 
                        return p[i].handler;

        log(LOG_ERR, "Unknow datalink type %d.\n", type);
        
        return NULL;
        
}



/*
 * Setup bpf rule 'bpf' for device identified by 'dev'.
 */
static int setup_bpf(capdev_t *dev, char *bpf)
{
        int ret;
	struct bpf_program fcode;

        if ( ! bpf )
                return 0;
        
        ret = pcap_compile(dev->pd, &fcode, bpf, 1, dev->netmask);
	if ( ret < 0 ) {
		log(LOG_ERR, "pcap: %s.\n", pcap_geterr(dev->pd));
                return -1;
	}

	ret = pcap_setfilter(dev->pd, &fcode);
        if ( ret < 0 ) {
		log(LOG_ERR, "pcap: %s.\n", pcap_geterr(dev->pd));
		return -1;
	}
        
	return 0;
}



/*
 * Setup pcap to capture from file 'filename'.
 */
static capdev_t *setup_capture_from_file(char *filename)
{
        capdev_t *dev;
	char error[PCAP_ERRBUF_SIZE];

        dev = malloc(sizeof(capdev_t));
        if ( ! dev ) {
                log(LOG_ERR, "memory exhausted.\n");
                return NULL;
        }
        
	dev->pd = pcap_open_offline(filename, error);
	if ( ! dev->pd ) {
                log(LOG_ERR, "pcap_open_offline: %s.\n", error);
                free(dev);
                return NULL;
	}

	return dev;
}



/*
 * Setup pcap to capture from device 'device'.
 */
static capdev_t *setup_capture_from_device(char *device)
{
        int ret;
        capdev_t *dev;
	char err[PCAP_ERRBUF_SIZE];
	bpf_u_int32 localnet, netmask;

        dev = malloc(sizeof(capdev_t));
        if ( ! dev ) {
                log(LOG_ERR, "memory exhausted.\n");
                return NULL;
        }
                
	dev->pd = pcap_open_live(device, config.snaplen, 1, 0, err);
	if ( ! dev->pd ) {
		log(LOG_ERR,"pcap_open_live: %s.\n", err);
                free(dev);
                return NULL;
	}

        ret = pcap_lookupnet(device, &localnet, &netmask, err);
	if ( ret < 0) {
                log(LOG_ERR,"pcap_lookupnet: %s.\n", err);
                pcap_close(dev->pd);
                free(dev);
                return NULL;
	}
        
	return dev;
}



/*
 * Poll our fd array for data,
 * return 0 on timeout, 1 on data availlable, -1 on error.
 */
static int do_poll(struct pollfd *pfd, unsigned int nfds, int timeout) 
{
        int ret;

        ret = poll(pfd, nfds, timeout);
        if ( ret < 0 ) {
                if ( errno == EINTR )
                        return 0;

                log(LOG_ERR, "poll returned an error.\n");
                return -1;
        }

        if ( ret == 0 )
                return 0;

        return 1;
}





/*
 * Start capturing from multiple devices,
 * only called if more than 1 capture device were configured.
 */
static int do_capture_from_multiple_devices(unsigned char *pcap_userdata) 
{
        int ret, i;
        capdev_t *dev;
        struct timeval tv;
        unsigned long last = 0;

        while (1) {
		ret = do_poll(pfd, interfaces_nbr, 1000);
                if ( ret == 0 ) {
                        wake_up_timer();
                        continue;
                }
                
                else if ( ret < 0 ) 
                        break;
                
		for (i = 0; i < interfaces_nbr; i++) {
                        dev = devices[i];

                        if ( pfd[i].revents & POLLIN ) {
                                errno = 0;
                                ret = pcap_dispatch(dev->pd, 1, dev->p_handler, pcap_userdata);
                                if ( ret < 0 && errno != EINTR )
                                        break;
                                
                                packet_counter++;
                        }
		}
                
                gettimeofday(&tv, NULL);
                if ( (tv.tv_sec - last) >= 1 ) {
                        last = tv.tv_sec;
                        wake_up_timer();
                }                
	}

        log(LOG_ERR, "packet capture interupted.\n");
        
        return ret;
}



/*
 * Start capturing from a single device.
 */
static int do_capture_from_single_device(unsigned char *pcap_userdata) 
{
        int ret;
        capdev_t *dev;
        long last = 0;
        struct timeval tv;

        dev = devices[0];
        
        while ( 1 ) {
                ret = do_poll(pfd, interfaces_nbr, 1000);
                if ( ret == 0 ) {
                        wake_up_timer();
                        continue;
                }

                else if ( ret < 0 )
                        break;

                if ( pfd[0].revents & POLLIN ) {
                        errno = 0;
                        
                        ret = pcap_dispatch(dev->pd, 1, dev->p_handler, pcap_userdata);
                        if ( ret < 0 ) {                        
                                if (  errno == EINTR ) 
                                        continue;
                                break;
                        }

                        packet_counter++;
                }
                
                gettimeofday(&tv, NULL);
                if ( (tv.tv_sec - last) >= 1 ) {
                        last = tv.tv_sec;
                        wake_up_timer();
                }
        }

        log(LOG_ERR,"pcap_loop: %s.\n", pcap_geterr(dev->pd));

        return ret;
}



/*
 * Start capturing from a single device.
 */
static int do_capture_from_file(unsigned char *pcap_userdata) 
{
        int ret;
        long last = 0;
        struct timeval tv;

        
        while ( 1 ) {
                errno = 0;
                ret = pcap_dispatch(file_dev->pd, 1, file_dev->p_handler, pcap_userdata);
                if ( ret < 0 ) {                        
                        if (  errno == EINTR ) 
                                continue;
                        break;
                }

                packet_counter++;
                
                gettimeofday(&tv, NULL);
                if ( (tv.tv_sec - last) >= 1 ) {
                        last = tv.tv_sec;
                        wake_up_timer();
                }
        }

        log(LOG_ERR,"pcap_loop: %s.\n", pcap_geterr(file_dev->pd));

        return ret;
}



/*
 * Gather stats from device 'pd'.
 */
static void pcap_dump_stat_from_device(pcap_t *pd) 
{
        int ret;
        struct pcap_stat stat;

        ret = pcap_stats(pd, &stat);
        if ( ret < 0 ) {
                log(LOG_ERR,"pcap_stats: %s.\n", pcap_geterr(pd));
                return;
        }

        log(LOG_INFO, "%u packets dropped by the kernel.\n\n", stat.ps_drop);
}



/*
 * Dump some process stats, gathered using getrusage.
 */
static void dump_system_stats(void) 
{
        struct rusage rusage;
	double total_time;
        
        getrusage(RUSAGE_SELF, &rusage);

        total_time = ((double) rusage.ru_utime.tv_usec +
                      (double) rusage.ru_stime.tv_usec) * 1e-6;
	total_time += ((double) rusage.ru_utime.tv_sec +
                       (double) rusage.ru_stime.tv_sec);
        
	log(LOG_INFO, "Average cpu time by packet : %fs, %fms, %fus.\n\n",
            (total_time / (double) packet_counter),
            (total_time / (double) packet_counter) * 1e3,
            (total_time / (double) packet_counter) * 1e6);
                       
        log(LOG_INFO, "Page reclaims = %ld\n", rusage.ru_minflt);
        log(LOG_INFO, "Page faults = %ld\n", rusage.ru_majflt);
        log(LOG_INFO, "Swap = %ld\n", rusage.ru_nswap);
}



/*
 * Add device identified by 'dev', to the devices and poll array.
 */
static int add_device(capdev_t *dev) 
{
        devices = realloc(devices, ++interfaces_nbr * sizeof(capdev_t *));
        if ( ! devices ) {
                log(LOG_ERR, "memory exhausted.\n");
                return -1;
        }

        
        pfd = realloc(pfd, interfaces_nbr * sizeof(struct pollfd));
        if ( ! pfd ) {
                log(LOG_ERR, "memory exhausted.\n");
                return -1;
        }
        
        devices[interfaces_nbr - 1] = dev;
        pfd[interfaces_nbr - 1].fd = dev->pcap_filedes;
        pfd[interfaces_nbr - 1].events = POLLIN;
        
        return 0;
}



/*
 * Shared setup capture function.
 */
static int capture_setup_generic(capdev_t *dev, char *bpf, int cap_from_file) 
{
        int ret;
        
        dev->p_handler = search_datalink_handler(pcap_datalink(dev->pd));
        if ( ! dev->p_handler )
                return -1;

        dev->bpf = bpf;
        ret = setup_bpf(dev, bpf);
        if ( ret < 0 )
                return -1;

        if ( ! cap_from_file ) {
                dev->pcap_filedes = pcap_fileno(dev->pd);
                ret = add_device(dev);
        } else 
                file_dev = dev;

        return ret;
}




/**
 * capture_from_device:
 * @device: The device to capture packet from.
 * @bpf: An optionnal Berkeley Packet Filter rule.
 *
 * Initialize the capture on specified device, with
 * an optionnal bpf rule.
 *
 * Returns: 0 on success, -1 otherwise.
 */
int capture_from_device(char *device, char *bpf) 
{
        capdev_t *dev;
        
        dev = setup_capture_from_device(device);
        if ( ! dev )
                return -1;

        return capture_setup_generic(dev, bpf, 0);
}




/**
 * capture_from_file:
 * @filename: The filename to read the packet from.
 * @bpf: An optionnal Berkeley Packet Filter rule.
 *
 * Initialize the capture on specified filename, with
 * an optionnal bpf rule.
 *
 * Returns: 0 on success, -1 otherwise.
 */
int capture_from_file(char *filename, char *bpf) 
{
        capdev_t *dev;

        dev = setup_capture_from_file(filename);
        if ( ! dev )
                return -1;

        return capture_setup_generic(dev, bpf, 1);
}




void capture_to_file(const char *filename) 
{
        capture_to = filename;
}




/**
 * capture_stats:
 *
 * Dump packet capture statistic.
 */
void capture_stats(void)
{
        int i;
        
        log(LOG_INFO, "%qd packets received by filter. (prelude counted),"
            " will reset after 2e64-1.\n", packet_counter);

        for (i = 0; i < interfaces_nbr; i++)
                pcap_dump_stat_from_device(devices[i]->pd);
        
        dump_system_stats();
}



/**
 * capture_setup_global_bpf:
 * @gbpf: A Berkeley Packet Filter rule.
 *
 * This function setup a BPF rule for all interfaces.
 *
 * Returns: 0 on success, -1 otherwise.
 */
int capture_setup_global_bpf(char *gbpf)
{
        int ret, i;
        char buf[1024];
        
        for ( i = 0; i < interfaces_nbr; i++ ) {
                
                if ( devices[i]->bpf ) {
                        snprintf(buf, sizeof(buf), "%s and %s", gbpf, devices[i]->bpf);
                        ret = setup_bpf(devices[i], buf);
                } else
                        ret = setup_bpf(devices[i], gbpf);
                
                if ( ret < 0 )
                        return -1;
        }
        
	return 0;
}




/**
 * capture_start:
 *
 * Start the capture on previously configured device.
 * This function will not return unless an error occur.
 *
 * Returns: -1 on error.
 */
int capture_start(void)
{
	int ret;
        unsigned char *pcap_userdata = NULL;

        decode_init(1000);

        if ( file_dev )
                ret = do_capture_from_file(pcap_userdata);
        
        else if ( interfaces_nbr > 1 )
                ret = do_capture_from_multiple_devices(pcap_userdata);

        else {
                if ( capture_to ) {
                        pcap_dumper_t *p;

                        p = pcap_dump_open(devices[0]->pd, capture_to);
                        if ( ! p ) {
                                log(LOG_ERR, "%s", pcap_geterr(devices[0]->pd));
                                return -1;
                        }

                        pcap_userdata = (unsigned char *) p;
                        devices[0]->p_handler = pcap_dump;
                }
                
                ret = do_capture_from_single_device(pcap_userdata);
        }
        
        return ret;
}




/**
 * capture_stop:
 *
 * Stop packet capture.
 */
void capture_stop(void)
{
        int i;
        
        for (i = 0; i < interfaces_nbr; i++)
                pcap_close(devices[i]->pd);
}
