/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		The IP fragmentation functionality.
 *		
 *
 * Authors:	Fred N. van Kempen <waltje@uWalt.NL.Mugnet.ORG>
 *		Alan Cox <Alan.Cox@linux.org>
 *
 * Fixes:
 *		Alan Cox	:	Split from ip.c , see ip_input.c for history.
 *		David S. Miller :	Begin massive cleanup...
 *		Andi Kleen	:	Add sysctls.
 *		xxxx		:	Overlapfrag bug.
 *		Ultima          :       ip_expire() kernel panic.
 *		Bill Hawes	:	Frag accounting and evictor fixes.
 *		John McDonald	:	0 length frag bug.
 *		Alexey Kuznetsov:	SMP races, threading, cleanup.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "common.h"
#include "packet.h"
#include "timer.h"
#include "ip_fragment.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "rqueue.h"
#include "hostdb.h"
#include "recycler.h"
#include "detect-plugins-api.h"


extern Packet_t *packet;


/*
 * "Fragment Offset" part
 */
#define IP_OFFSET       0x1FFF

typedef struct frag_item_t
{
        int offset; /* offset of frag in Ip datagram */
        int end;    /* last byte of data in datagram */
        int len;    /* lenght of this fragment       */
        unsigned char *data;
        struct frag_item_t *next;
        struct frag_item_t *prev;
} frag_item_t;


/*
 * Describe an entry in the "incomplete datagrams" queue.
 */
struct ipq {
        Packet_t *packet;
        struct __iphdr *ip;
	uint8_t		last_in;
        
#define FIRST_IN		2
#define LAST_IN			1

	frag_item_t	*fragments;	/* linked list of received fragments */
	int		len;		/* total length of original datagram */
	int		meat;
	timer_t         timer;	        /* when will this queue expire?	     */
};


static const int my_id = 1;


/*
 * Destruction primitives.
 */

/*
 * Complete destruction of ipq.
 */
static void ip_frag_destroy(struct ipq *qp)
{
	frag_item_t *fp;
        
	/*
         * Release all fragment data.
         */
	fp = qp->fragments;
	while (fp) {
                frag_item_t *next = fp->next;

                free(fp->data);
                free(fp);

                fp = next;
	}

	/*
         * Finally, release the queue descriptor itself.
         */
        free(qp);
}



/*
 * Kill ipq entry.
 */
static __inline__ void ipq_kill(hostdb_t *hdb, struct ipq *ipq)
{
        packet_release(ipq->packet);
        timer_destroy(&ipq->timer);
        hostdb_del(hdb, my_id);
        ip_frag_destroy(ipq);
}



/*
 * Oops, a fragment queue timed out.  Kill it.
 */
static void ip_expire(unsigned long arg)
{
        struct ipq *qp;
        hostdb_t *hdb = (hostdb_t *) arg;

        qp = (struct ipq *) hostdb_get_plugin_data(hdb, my_id);

        if( ! qp->fragments )
                prelude_do_report(qp->packet, maybe_faked, "Possible Ip-expire attack",
                                  "A packet with the fragment bit turned on was received\n "
                                  "but no fragment were sent by this host.\n " 
                                  "This could be issued in order to operate a "
                                  "Denial Of Services attack against your "
                                  "machine.");
        
        ipq_kill(hdb, qp);
}



/*
 * Add an entry to the 'ipq' queue for a newly received IP datagram.
 */
static struct ipq *ip_frag_create(hostdb_t *hdb, struct __iphdr *iph)
{
	struct ipq *qp;

        qp = malloc(sizeof(struct ipq));
        if ( ! qp ) {
                log(LOG_ERR,"ip_frag_create(): no memory left.\n");
                return NULL;
        }

        packet_lock(packet);
        
        qp->ip = iph;
        qp->packet = packet;
        qp->len = 0;
        qp->last_in = 0;
	qp->meat = 0;
	qp->fragments = NULL;
        
	/*
         * Initialize a timer for this entry.
         */
        timer_init(&qp->timer);
        timer_set_data(&qp->timer, hdb);
        timer_set_expire(&qp->timer, 1);
        timer_set_callback(&qp->timer, ip_expire);
                
	return qp;
}


/*
 * Add new segment to existing queue.
 */
static void ip_frag_queue(struct ipq *qp, const struct __iphdr *ip, frag_item_t *frag)
{
        const struct ip *iph = &ip->ip_hdr;
	frag_item_t *prev, *next;
	int flags, offset;
	int ihl, end;
                
	offset = ntohs(iph->ip_off);
        
        flags = offset & ~IP_OFFSET;
	offset &= IP_OFFSET;
	offset <<= 3;		/* offset is in 8-byte chunks */
	ihl = iph->ip_hl * 4;

        frag->offset = offset;
        
	/*
         * Determine the position of this fragment.
         */
	end = offset + (ntohs(iph->ip_len) - ihl);
        
	/*
         * Is this the final fragment ?
         */
	if ((flags & IP_MF) == 0) {
                /*
                 * If we already have some bits beyond end
		 * or have different end, the segment is corrrupted.
		 */
		if (end < qp->len ) {
                        prelude_do_report(packet, maybe_faked, "Ip Defragmentation attack",
                                          "Received last fragment of a fragmented stream "
                                          "with end of data at byte %d. \n"
                                          "But current received data len (%d) is beyond that end.\n "
                                          "This could be issued in order to operate a "
                                          "Denial Of Services attack against your machine.\n",
                                          end, qp->len);

                        goto err;
                }

                if ( (qp->last_in & LAST_IN) && end != qp->len ) {
                        prelude_do_report(packet, maybe_faked, "Ip Defragmentation attack",
                                          "Received last fragment of a fragmented stream "
                                          "with end of data at byte %d. \n"
                                          "Then received another last fragment for the same stream.\n "
                                          "with end of data == %d != to the first end of data gathered (%d).\n\n "
                                          "This could be issued in order to operate a "
                                          "Denial Of Services attack against your machine.\n ",
                                          qp->len, end, qp->len);
                        goto err;
                }
                
		qp->last_in |= LAST_IN;
		qp->len = end;
	} else {
		if (end & 7) 
			end &= ~7;
		
		if (end > qp->len) {
			/*
                         * Some bits beyond end -> corruption.
                         */
			if (qp->last_in & LAST_IN) {
                                prelude_do_report(packet, maybe_faked,
                                                  "Ip Defragmentation attack",
                                                  "Received last fragment of a fragmented stream "
                                                  "with end of data at %d bytes.\n "
                                                  "Then received another packet with MF flag set "
                                                  "and end of data at %d bytes (beyond the end).\n\n "
                                                  "This could be issued in order to operate a "
                                                  "Denial Of Services attack against your machine.\n",
                                                  qp->len, end);
                                goto err;
                        }
                        
			qp->len = end;
		}
	}
        
	if (end == offset) 
		goto err;
        
	
	/*
         * Find out which fragments are in front and at the back of us
	 * in the chain of fragments so far.  We must know where to put
	 * this fragment, right?
	 */
	prev = NULL;
	for (next = qp->fragments; next != NULL; next = next->next) {
                if ( next->offset >= offset)
			break;	/* bingo! */
		prev = next;
	}

	/*
         * We found where to put this one.  Check for overlap with
	 * preceding fragment, and, if needed, align things so that
	 * any overlaps are eliminated.
	 */
	if (prev) {
		int i = (prev->offset + prev->len) - offset;

		if (i > 0) {
			offset += i;
			if (end <= offset)
				goto err;
		}
	}

	while (next && next->offset < end) {
		int i = end - next->offset; /* overlap is 'i' bytes */
                
		if (i < next->len) {
			/*
                         * Eat head of the next overlapped fragment
			 * and leave the loop. The next ones cannot overlap.
			 */
			next->offset += i;
			qp->meat -= i;
                        break;
		} else {
			frag_item_t *free_it = next;

			/*
                         * Old fragment is completely overridden with
			 * new one drop it.
			 */
			next = next->next;

			if (prev)
				prev->next = next;
			else
				qp->fragments = next;

			qp->meat -= free_it->len;

                        free(free_it->data);
                        free(free_it);
		}
	}

        /*
         * Insert this fragment in the chain of fragments.
         */
              
	frag->next = next;
	if (prev)
		prev->next = frag;
	else
		qp->fragments = frag;

        qp->meat += frag->len;
        if (offset == 0) 
		qp->last_in |= FIRST_IN;
        
        return;

 err:
        
        free(frag->data);
        free(frag);	
}


/*
 * Build a new IP datagram from all its fragments.
 *
 * FIXME: We copy here because we lack an effective way of handling lists
 * of bits on input. Until the new skb data handling is in I'm not going
 * to touch this with a bargepole. 
 */
static int ip_frag_reasm(hostdb_t *hdb, struct ipq *qp, unsigned char **dp)
{
	struct __iphdr *ip = qp->ip;
        struct ip *iph = &ip->ip_hdr;
        frag_item_t *fp, *bkp;
        int len;
	int ihlen;
        unsigned char *p;
        
        /*
         * Allocate a new buffer for the datagram.
         */

        ihlen = iph->ip_hl * 4;
	len = ihlen + qp->len;

	if (len > 65535) {
                prelude_do_report(packet, maybe_faked, "Oversized packet",
                                  "Oversized packet ( > 65535) while doing reassembly.\n\n "
                                  "\n\nThis could be issued in order to "
                                  "operate a Denial of Services attack"
                                  "against your machine.");
                ipq_kill(hdb, qp);
                return -1;
        }
        
        *dp = p = (unsigned char *) malloc(len);
	if ( ! p ) {
                log(LOG_ERR,"malloc(%d).\n", len);
                ipq_kill(hdb, qp);
		return -1;
        }

        packet_release(qp->packet);
        hostdb_del(hdb, my_id);
        timer_destroy(&qp->timer);

        /*
         * Fixup the new IP headers
         */
        iph->ip_off = 0;
        iph->ip_len = htons(len);
        
        /*
         * Copy the original IP headers into the new buffer.
         */
        memcpy(p, iph, ihlen);
        p += ihlen;
        
	/*
         * Copy the data portions of all fragments into the new buffer.
         */
	for (fp = qp->fragments; fp; fp = bkp) {
                
                memcpy(p, fp->data, fp->len);
                p += fp->len;
                
                free(fp->data);

                bkp = fp->next;
                free(fp);
        }

        free(qp);
        
	return len;
}

static frag_item_t *nfrag(const unsigned char *data, int len) 
{
        frag_item_t *f = NULL;

        f = (frag_item_t *) malloc(sizeof(frag_item_t));
        if ( ! f ) {
                log(LOG_ERR,"malloc(%d).\n", sizeof(frag_item_t));
                return NULL;
        }

        f->len = len;

        f->data = (unsigned char *) malloc(len);
        if ( ! f->data ) {
                log(LOG_ERR,"malloc(%d).\n", len);
                free(f);
                return NULL;
        }
        
        memcpy(f->data, data, len);
        
        f->next = NULL;
        f->prev = NULL;
        
        return f;
}

/* Process an incoming IP datagram fragment. */
int ip_defrag(struct __iphdr *iph, const unsigned char *data,
              int len, unsigned char **dp)
{
        struct ipq *qp;
        hostdb_t *hdb;
        frag_item_t *f;
        
        hdb = hostdb_search(&iph->ip_hdr);
        if ( ! hdb ) {
                hdb = hostdb_new(&iph->ip_hdr);
                if (! hdb )
                        return -1;

                qp = NULL;
        } else
                qp = (struct ipq *) hostdb_get_plugin_data(hdb, my_id);
                
        if ( ! qp ) {
                qp = ip_frag_create(hdb, iph);
                if ( ! qp ) {
                        hostdb_del(hdb, my_id);
                        return -1;
                }
                
                hostdb_set_plugin_data(hdb, my_id, (unsigned long)qp);
        }
        
        
        /*
         * Lookup (or create) queue header
         */
        f = nfrag(data, len);        
        if ( ! f ) {
                ipq_kill(hdb, qp);
                return -1;
        }
                
        ip_frag_queue(qp, iph, f);

        if (qp->last_in == (FIRST_IN|LAST_IN) && qp->meat == qp->len)
                return ip_frag_reasm(hdb, qp, dp);
   
        return -1;
}




