/*****
*
* Copyright (C) 1999,2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

/*
 * extract.h stolen from tcpdump source
 */
#include "extract.h"
#include "common.h"
#include "packet.h"
#include "pluginlib.h"
#include "optparse.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "timer.h"
#include "rqueue.h"



static Packet_t *pkt;
static const char *proto;



/*
 * From rfc793 :
 * Options may occupy space at the end of the TCP header and are a
 * multiple of 8 bits in length.  All options are included in the
 * checksum.  An option may begin on any octet boundary.  There are two
 * cases for the format of an option:
 *
 *  Case 1:  A single octet of option-kind.
 *
 *  Case 2:  An octet of option-kind, an octet of option-length, and
 *           the actual option-data octets.
 *
 * The option-length counts the two octets of option-kind and
 * option-length as well as the option-data octets.
 *
 * Note that the list of options may be shorter than the data offset
 * field might imply.  The content of the header beyond the
 * End-of-Option option must be header padding (i.e., zero).
 *
 * A TCP must implement all options.
 */



/*
 * Verify if the option 'opt' is one of
 * the 1 byte only option (nop || eol).
 */
static int is_1byte_option(int opt) 
{
        if ( opt == __TCPOPT_NOP ) {
                return 0;
        }

        else if (opt == __TCPOPT_EOL) {
                return 0;
        }

        return -1;
}



/*
 * Verify that an option is valid :
 * - verify that this option len is > 2 bytes.
 * - verify that this option len is < than our total option len.
 * - do some bound check on our option buffer, to avoid going out of bound.
 */
static int is_option_valid(const unsigned char *optbuf, int optlen, int totlen) 
{        
        if ( optlen < 2 ) {
                prelude_do_report(pkt, maybe_faked,
                                  "Options not valid",
                                  "%s options not valid : options is not \"nop\" or \"eol\" "
                                  "so option len (%d) should be >= 2.\n\n "
                                  "This could be issued in order to make your operating "
                                  "system to handle this packet in an unpredictable way\n "
                                  "or as a DOS attack.",
                                  proto, optlen);
                return -1;
        }
        
        if ( optlen > totlen ) {
                prelude_do_report(pkt, maybe_faked,
                                  "Options not valid",
                                  "%s options not valid : option len (%d) is > remaining "
                                  "total options len (%d).\n\n "
                                  "This could be issued in order to make your operating system\n "
                                  "to handle this packet in an unpredictable way\n "
                                  "or as a DOS attack.",
                                  proto, optlen, totlen);
                return -1;
        }

        /*
         * This check should never be reached because
         * of the optlen > totlen test... use an assert.
         */
        assert( (optbuf + (optlen - 2)) <= (optbuf + (totlen - 2)) );

        return 0;
}



/*
 * Verify that our total options len is big enough
 * to contain a len byte, which mean totlen must be
 * >= 2 (1 byte for optkind, and 1 for optlen).
 */
static int is_len_byte_ok(int totlen) 
{
        if ( totlen < 2 ) {
                prelude_do_report(pkt, maybe_faked,
                                  "Options not valid",
                                  "%s options not valid : options is not \"nop\" or \"eol\" "
                                  "so total options len (%d) should be >= 2 (1 byte kind, 1 "
                                  "byte len).\n\n "
                                  "This could be issued in order to make your operating "
                                  "system to handle this packet in an unpredictable way\n "
                                  "or as a DOS attack.",
                                  proto, totlen);
                return -1;
        }
        
        return 0;
}



/*
 * Walk options in 'optbuf' of total len 'totlen'.
 * Return the number of byte parsed before error.
 */
static int walk_options(const unsigned char *optbuf, int totlen) 
{
        int origlen = totlen;
        int opt, optlen, ret;
        
        do {
                opt = *optbuf++;

                if ( is_1byte_option(opt) == 0 )
                        totlen -= 1;
                else {
                        if ( is_len_byte_ok(totlen) < 0 )
                                return origlen - totlen;

                        optlen = *optbuf++;
                        
                        ret = is_option_valid(optbuf, optlen, totlen);
                        if ( ret < 0 )
                                return origlen - (totlen - 2);
                        
                        totlen -= optlen;
                        optbuf += optlen - 2;

                }
                
                assert(totlen >= 0);
        } while ( totlen != 0 );

        return origlen - totlen;
}



/*
 *
 */
int tcp_optdump(Packet_t *packet, const unsigned char *optbuf, unsigned int totlen) 
{
        pkt = packet;
        proto = "tcp";
        
        if ( totlen > __MAX_OPTS_LEN ) {
                prelude_do_report(pkt, maybe_faked,
                                  "Options not valid",
                                  "%s options not valid : total option len (%d) > maximum "
                                  "option len (%d)\n\n "
                                  "This could be issued in order to make your operating "
                                  "system to handle this packet in an unpredictable way\n "
                                  "or as a DOS attack.",
                                  proto, totlen, __MAX_OPTS_LEN);
                return -1;
        }
        
        return walk_options(optbuf, totlen);
}



/*
 *
 */
int ip_optdump(Packet_t *packet, const unsigned char *optbuf, unsigned int totlen) 
{
        pkt = packet;
        proto = "ip";
        
        if ( totlen > __MAX_OPTS_LEN ) {
                prelude_do_report(pkt, maybe_faked,
                                  "Options not valid",
                                  "%s options not valid : total option len (%d) > maximum "
                                  "option len (%d)\n\n "
                                  "This could be issued in order to make your operating "
                                  "system to handle this packet in an unpredictable way\n "
                                  "or as a DOS attack.",
                                  proto, totlen, __MAX_OPTS_LEN);
                return -1;
        }
        
        return walk_options(optbuf, totlen);
}














