/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <pcap.h>
#include <ctype.h>

#include "common.h"
#include "list.h"
#include "packet.h"
#include "ip_fragment.h"
#include "packet-decode.h"
#include "optparse.h"
#include "recycler.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "timer.h"
#include "rqueue.h"
#include "detect-plugins.h"

extern struct list_head __tcp_plugins;
extern struct list_head __udp_plugins;
extern struct list_head __ip_plugins;
extern struct list_head __ipencap_plugins;
extern struct list_head __data_plugins;
extern struct list_head __icmp_plugins;
extern struct list_head __igmp_plugins;
extern struct list_head __arp_plugins;
extern struct list_head __rarp_plugins;
extern struct list_head __ether_plugins;
extern struct list_head __all_plugins;

static int caplen;
static const struct pcap_pkthdr *pcap_hdr;
static const unsigned char *snapend;
static unsigned char *frag_ptr = NULL;
int depth;
Packet_t *packet;

/*
 * recyclers.
 */
static recycler_t *packet_recycler;
static recycler_t *data_recycler;
static recycler_t *ip_recycler;
static recycler_t *tcp_recycler;
static recycler_t *udp_recycler;
static recycler_t *ip_recycler;
static recycler_t *ether_recycler;
static recycler_t *arp_recycler;
static recycler_t *rarp_recycler;
static recycler_t *igmp_recycler;
static recycler_t *icmp_recycler;


#define _get_chunk(recycler, ptr) do {             \
        ptr = RecyclerGetChunk(recycler);          \
        if ( ! ptr ) {                             \
           log(LOG_ERR, "can't slice packet... no more free chunk.\n"); \
           return;                                 \
        }                                          \
                                                   \
        if ( RecyclerIsLocked(packet) == TRUE )    \
           RecyclerLockChunk(ptr);                 \
} while (0)



#define incr_depth()                          \
        if ( verify_depth(__FUNCTION__) < 0 ) \
                return;


static int verify_depth(const char *function) 
{
        /*
         * go to next depth.
         */
        depth++;

        /*
         * depth + 1 must be availlable.
         */
        if ( (depth + 1) < MAX_PKTDEPTH ) {
                packet[depth + 1].proto = p_end;
                return 0;
        }

        /*
         * if it is not, thing goes wrong.
         */
        packet[depth].proto = p_end;
        prelude_do_report(packet, maybe_faked,
                          "Packet depth is too high",
                          "%s : Maximum packet depth (%d) reached.\n",
                          function, MAX_PKTDEPTH);

        return -1;
}




static void SliceAndStoreDataPkt(const u_char *p)
{
        PktData_t *data;
        
        /*
         * Don't issue warning if there is no data,
         * it can be absolutly normal.
         */
        if ( !frag_ptr && p + 1 > snapend )
                return;

        incr_depth();
        
        packet[depth].len = sizeof(PktData_t) + caplen;
        packet[depth].proto = p_data;
        _get_chunk(data_recycler, data);
        packet[depth].p.pkt_data = data;

        data->size = caplen;
        data->data[0] = 0;

        memcpy(data->data, p, caplen);
        data->data[caplen] = '\0';
     
	detect_plugins_run(packet, &__data_plugins, "data");

        data->data[0] = 0;
}

static void SliceAndStoreTcpPkt(const u_char *p)
{
        struct __tcphdr *tcp;
        int hlen;
        
        incr_depth();
        
        if ( caplen < sizeof(struct tcphdr) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated TCP header",
                                  "truncated TCP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct tcphdr), caplen);
                return;
        }

        packet[depth].len = sizeof(struct __tcphdr);
        packet[depth].proto = p_tcp;
        _get_chunk(tcp_recycler, tcp);
        packet[depth].p.tcp = tcp;
        
        memcpy(&tcp->tcp_hdr, p, sizeof(struct tcphdr));        
        
        hlen = tcp->tcp_hdr.th_off * 4;
        if ( hlen > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "bad TCP header length",
                                  "bad TCP header length (hlen=%d, caplen=%d).\n\n ",
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", hlen, caplen);
                return;
        }
        
        caplen -= hlen;
        tcp->opts_len = hlen - sizeof(struct tcphdr);
        if ( tcp->opts_len > 0 ) {
                const unsigned char *ptr = p + sizeof(struct tcphdr);
                
                tcp->opts_len = tcp_optdump(packet, ptr, tcp->opts_len);
                if ( tcp->opts_len > 0 )
                        memcpy(tcp->opts, ptr, tcp->opts_len);
        }

	detect_plugins_run(packet, &__tcp_plugins, "tcp");
        
        SliceAndStoreDataPkt(p + hlen);
}

static void SliceAndStoreUdpPkt(const u_char *p)
{
        struct udphdr *udp;
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct udphdr) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated UDP",
                                  "truncated UDP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct udphdr), caplen);
                 return;
        }

        incr_depth();
        caplen -= sizeof(struct udphdr);
        /*
         * done.
         */
        packet[depth].len = sizeof(struct  udphdr);
	packet[depth].proto = p_udp;
        _get_chunk(udp_recycler, udp);
        packet[depth].p.udp_hdr = udp;
        
	memcpy(udp, p, sizeof(struct udphdr));

	detect_plugins_run(packet, &__udp_plugins, "udp");
        
	SliceAndStoreDataPkt(p + sizeof(struct udphdr));
}



static void SliceAndStoreIcmpPkt(const unsigned char *p)
{
        int len;
        struct icmp *hdr;
        uint8_t type_tbl[] = { 8,  /* type 0 */
                               0,  /* type 1 */
                               0,  /* type 2 */
                               36, /* type 3 */
                               36, /* type 4 */
                               36, /* type 5 */
                               0,  /* type 6 */
                               0,  /* type 7 */
                               8,  /* type 8 */
                               0,  /* type 9 router advertissement message
                                    * require extenssible struct */

                               8,  /* type 10 */
                               36, /* type 11 */
                               36, /* type 12 */
                               20, /* type 13 */
                               20, /* type 14 */
                               8,  /* type 15 */
                               8,  /* type 16 */
                               12, /* type 17 address mask request */
                               12, /* type 18 address mask reply   */
        };


        if ( caplen < ICMP_MINLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ICMP",
                                  "truncated ICMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", ICMP_MINLEN, caplen);
                return;
        }

        _get_chunk(icmp_recycler, hdr);
        memcpy(hdr, p, ICMP_MINLEN);
        
        if ( hdr->icmp_type < sizeof(type_tbl) )
                len = type_tbl[hdr->icmp_type];
        else
                len = ICMP_MINLEN;

        if ( caplen < len ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ICMP",
                                  "truncated ICMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", len, caplen);
                return;
        }

        caplen -= len;
        packet[depth].len = len;
        packet[depth].proto = p_icmp;
        packet[depth].p.icmp_hdr = hdr;
        
        detect_plugins_run(packet, &__icmp_plugins, "icmp");
        SliceAndStoreDataPkt(p + len);
}


static void SliceAndStoreIgmpPkt(const u_char *p)
{
        struct igmp *igmp;
        
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct igmp) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IGMP",
                                  "truncated IGMP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct igmp), caplen);
                return;
        }

        incr_depth();
        caplen -= sizeof(struct igmp);
        /*
         * done.
         */
        
        packet[depth].len = sizeof(struct igmp);
	packet[depth].proto = p_igmp;
        _get_chunk(igmp_recycler, igmp);
        packet[depth].p.igmp_hdr = igmp;
	memcpy(igmp, p, sizeof(struct igmp));

	detect_plugins_run(packet, &__igmp_plugins, "igmp");
             
        SliceAndStoreDataPkt(p + sizeof(struct igmp));
}

static void SliceAndStoreArpPkt(const u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct ether_arp) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ARP",
                                  "truncated ARP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct ether_arp), caplen);
                return;
        }

        incr_depth();
        caplen -= sizeof(struct ether_arp);
        /*
         * done.
         */

        packet[depth].len = sizeof(struct ether_arp);
	packet[depth].proto = p_arp;
        _get_chunk(arp_recycler, packet[depth].p.arp_hdr);
        memcpy(packet[depth].p.arp_hdr, p, sizeof(struct ether_arp));

	detect_plugins_run(packet, &__arp_plugins, "arp");
           
        SliceAndStoreDataPkt(p + sizeof(struct ether_arp));
}

static void SliceAndStoreRarpPkt(const u_char *p)
{
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct ether_arp) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated RARP",
                                  "truncated RARP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct ether_arp), caplen);
                return;
        }

        incr_depth();
        caplen -= sizeof(struct ether_arp);
        /*
         * Done.
         */

        packet[depth].len = sizeof(struct ether_arp);
	packet[depth].proto = p_rarp;
        _get_chunk(rarp_recycler, packet[depth].p.arp_hdr);
	memcpy(packet[depth].p.arp_hdr, p, sizeof(struct ether_arp));

	detect_plugins_run(packet, &__rarp_plugins, "rarp");
          
        SliceAndStoreDataPkt(p + sizeof(struct ether_arp));
}



/*
 * Return -1 on error,
 * 1 if a packet was reassembled.
 * 0 if a fragment was queued.
 */

#define FRAGMENT_QUEUED 0
#define PACKET_REASM 1

static int handle_ip_fragment(const u_char *p, struct __iphdr *ip,
                              unsigned int *hlen, unsigned int *len) 
{
        int ret;
                
        if ( ((long) (*len - *hlen)) <= 0 ) {
                prelude_do_report(packet, maybe_faked,
                                  "Invalid fragment data size (too short)",
                                  "Caught a fragment with total len ((len=%d - hlen=%d) = %d)\n "
                                  "caplen is %d, but the fragment size is too short.\n",
                                  *len, *hlen, *len - *hlen, caplen);
                return -1;
        }
                
        if ( ((long) (*len - *hlen)) > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "Invalid fragment data size (too big)",
                                  "Caught a fragment with total len ((len=%d - hlen=%d) = %d)\n"
                                  "bigger than captured len (%d).\n",
                                  *len, *hlen, *len - *hlen, caplen);
                return -1;
        }
                
        ret = ip_defrag(ip, p + sizeof(struct ip), *len - *hlen, &frag_ptr);
        if ( ret > 0 ) {            
                memcpy(&ip->ip_hdr, frag_ptr, sizeof(struct ip));
                caplen = ret;
                *hlen = ip->ip_hdr.ip_hl * 4;
                *len = ntohs(ip->ip_hdr.ip_len);
                
                if ( *hlen > caplen ) {
                        prelude_do_report(packet, maybe_faked,
                                          "Invalid reassembled header len", 
                                          "bad defragmented IP header length (hlen=%d, caplen=%d).\n\n ",
                                          "This could be used as a DOS in order to make your kernel IP stack\n "
                                          "or your network applications crash.", hlen, caplen);
                        free(frag_ptr);
                        frag_ptr = NULL;
                        return -1;
                }

                return PACKET_REASM;
        }

        return FRAGMENT_QUEUED;
}




static void SliceAndStoreIpPkt(const u_char *p, proto_enum_t ip_kind)
{
        unsigned int off, len, hlen;
        struct ip *ip_hdr;
        struct __iphdr *ip;
        
        incr_depth();
        
        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct ip) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IP",
                                  "truncated IP (ip header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct ip), caplen);
                return;
        }
        
        /*
         * End check.
         */

        packet[depth].len = sizeof(struct __iphdr);
        packet[depth].proto = ip_kind;
        _get_chunk(ip_recycler, ip);
        packet[depth].p.ip = ip;
        ip_hdr = &ip->ip_hdr;
        
        memcpy(ip_hdr, p, sizeof(struct ip));

        /*
         * This is only a check in case DLT_RAW is used.
         */
        if ( ip_hdr->ip_v != 4 )
                return;
        
        off = ntohs(ip_hdr->ip_off);
        len = ntohs(ip_hdr->ip_len);
        hlen = ip_hdr->ip_hl * 4;
                
        if ( !(off & 0x3fff) && caplen < len ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IP",
                                  "truncated IP ( len=%d caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", len, caplen);
        }

        
        if ( hlen > caplen ) {
                prelude_do_report(packet, maybe_faked,
                                  "bad IP header length",
                                  "bad IP header length (hlen=%d, caplen=%d).\n\n ",
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", hlen, caplen);
                return;
        }

        ip->opts_len = hlen - sizeof(struct ip);
        if ( ip->opts_len > 0 ) {
                const unsigned char *ptr = p + sizeof(struct ip);
                
                ip->opts_len = ip_optdump(packet, ptr, ip->opts_len);
                if ( ip->opts_len > 0 )
                        memcpy(ip->opts, ptr, ip->opts_len);
        }
        
        off = ntohs(ip_hdr->ip_off);
        if ( off & 0x3fff ) {
                int ret;

                ret = handle_ip_fragment(p, ip, &hlen, &len);
                if ( ret == PACKET_REASM )
                        p = frag_ptr;

                else if ( ret == FRAGMENT_QUEUED ) {
                        SliceAndStoreDataPkt(p + hlen);
                        return;
                }
                
                else
                        return;
        }
        
        caplen -= hlen;

        if ( ip_kind == p_ip )
                detect_plugins_run(packet, &__ip_plugins, "ip");
        else
                detect_plugins_run(packet, &__ipencap_plugins, "ipencap");

     	switch (ip_hdr->ip_p) {
	case IPPROTO_IP:
                /* Ip encapsulated in Ip */
                SliceAndStoreIpPkt(p + hlen, p_ipencap);
		break;
	case IPPROTO_TCP:
                SliceAndStoreTcpPkt(p + hlen);
		break;
	case IPPROTO_UDP: 
                SliceAndStoreUdpPkt(p + hlen);
		break;
	case IPPROTO_ICMP:
                SliceAndStoreIcmpPkt(p + hlen);
		break;
	case IPPROTO_IGMP:
                SliceAndStoreIgmpPkt(p + hlen);
		break;
	default:
                SliceAndStoreDataPkt(p + hlen);
		break;
	}
}



__inline__ static void switch_ethertype(unsigned short ethertype, const unsigned char *p) 
{
        switch (ethertype) {
	case ETHERTYPE_IP:
                SliceAndStoreIpPkt(p, p_ip);
		break;
	case ETHERTYPE_ARP:
                SliceAndStoreArpPkt(p);
		break;
	case ETHERTYPE_REVARP:
                SliceAndStoreRarpPkt(p);
		break;
	default:
                SliceAndStoreDataPkt(p);
                break;
	}
}




void SliceAndStoreEtherPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        struct ether_header *hdr;
                
        _get_chunk(packet_recycler, packet);
                
        depth = -1;
        incr_depth();

        packet[depth].len = sizeof(struct ether_header);
        packet[depth].proto = p_ether;
        _get_chunk(ether_recycler, hdr);
        packet[depth].p.ether_hdr = hdr;
               
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;

        /*
         * Sanity check.
         */
        if ( caplen < sizeof(struct ether_header) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ETHER",
                                  "truncated ETHER (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.",
                                  sizeof(struct ether_header), caplen);
                return;
        }

        caplen -= sizeof(struct ether_header);

        /*
         * Ok.
         */
	memcpy(hdr, p, sizeof(struct ether_header));

	detect_plugins_run(packet, &__ether_plugins, "ether");

	switch_ethertype(ntohs(hdr->ether_type), p + sizeof(struct ether_header));
        
        detect_plugins_run(packet, &__all_plugins, "all");

        if ( frag_ptr ) {
                free(frag_ptr);
                frag_ptr = NULL;
        }
}

#define    NULL_HDRLEN 4

void SliceAndStoreNullPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        _get_chunk(packet_recycler, packet);      
        depth = -1;
          
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        /*
         * Sanity check.
         */
        if ( caplen < NULL_HDRLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated NULL",
                                  "truncated NULL (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", NULL_HDRLEN, caplen);
                return;
        }

        caplen -= NULL_HDRLEN;  
        SliceAndStoreIpPkt(p + NULL_HDRLEN, p_ip);
}

#define PPP_HDRLEN 4

void SliceAndStorePppPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        _get_chunk(packet_recycler, packet);        
        depth = -1;
            
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        /*
         * Sanity check.
         */
        if ( caplen < PPP_HDRLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated PPP",
                                  "truncated PPP (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", PPP_HDRLEN, caplen);
                return;
        }

        caplen -= PPP_HDRLEN;
        
        SliceAndStoreIpPkt(p + PPP_HDRLEN, p_ip);
}

#define PPP_BSDI_HDRLEN 24

void SliceAndStorePppBsdosPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        _get_chunk(packet_recycler, packet);        
        depth = -1;
        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        /*
         * Sanity check.
         */
        if ( caplen < PPP_BSDI_HDRLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated PPP(bsdos)",
                                  "truncated PPP(bsdos) (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", PPP_BSDI_HDRLEN, caplen);
                return;
        }

        caplen -= PPP_BSDI_HDRLEN;
        
        SliceAndStoreIpPkt(p + PPP_BSDI_HDRLEN, p_ip);
}



void SliceAndStoreFddiPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        _get_chunk(packet_recycler, packet);
        depth = -1;
            
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        /*
         * Sanity check.
         */
        if ( caplen < FDDI_HDRLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated FDDI",
                                  "truncated FDDI (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", FDDI_HDRLEN, caplen);
                return;
        }        
        
        caplen -= FDDI_HDRLEN;
        SliceAndStoreIpPkt(p + FDDI_HDRLEN, p_ip);
}



#define ATM_HDRLEN 8
void SliceAndStoreAtmPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        unsigned short ethertype;

        _get_chunk(packet_recycler, packet);
        depth = -1;
        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        
        /*
         * Sanity check.
         */
        if ( caplen < ATM_HDRLEN ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated ATM",
                                  "truncated ATM (header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", ATM_HDRLEN, caplen);
                return;
        }        

        if (p[0] != 0xaa || p[1] != 0xaa || p[2] != 0x03) {
                p += 20;
                caplen -= 20;
        }

        ethertype = p[6] << 8 | p[7];
        caplen -= ATM_HDRLEN;

        switch_ethertype(ethertype, p + ATM_HDRLEN);
        
        SliceAndStoreIpPkt(p + ATM_HDRLEN, p_ip);
}



void SliceAndStoreRawPkt(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
{
        struct ip ip;
        
        pcap_hdr = h;
        snapend = p + h->caplen;
        caplen = h->caplen;
        depth = -1;

        if ( caplen < sizeof(struct ip) ) {
                prelude_do_report(packet, maybe_faked,
                                  "truncated IP",
                                  "truncated IP (ip header len=%d > caplen=%d ).\n\n "
                                  "This could be used as a DOS in order to make your kernel IP stack\n "
                                  "or your network applications crash.", sizeof(struct ip), caplen);
                return;
        }

        memcpy(&ip, p, sizeof(struct ip));
        
        switch(ip.ip_v) {

        case 4:
                SliceAndStoreIpPkt(p, p_ip);
                break;

        case 6:
                log(LOG_ERR, "we don't handle IPV6 yet.\n");
                break;
        }
}




int decode_init(int elements) 
{
        log(LOG_INFO, "\tCreating %d elements for Packets (%d bytes).\n",
            elements, elements * (20 * sizeof(Packet_t)));

        
        packet_recycler = RecyclerCreate(20 * sizeof(Packet_t), elements);
        if (! packet_recycler ) {
                log(LOG_ERR,"RecyclerCreate().\n");
                return -1;
        }


        tcp_recycler = RecyclerCreate(sizeof(struct __tcphdr), elements);
        udp_recycler = RecyclerCreate(sizeof(struct udphdr), elements);
        icmp_recycler = RecyclerCreate(sizeof(struct icmp), elements);
        igmp_recycler = RecyclerCreate(sizeof(struct igmp), elements);
        arp_recycler = RecyclerCreate(sizeof(struct ether_arp), elements);
        rarp_recycler = RecyclerCreate(sizeof(struct ether_arp), elements);
        ether_recycler = RecyclerCreate(sizeof(struct ether_header), elements);

        /*
         * Max data size of an IP defragmented packet will be 65535 bytes.
         */
        data_recycler = RecyclerCreate(sizeof(PktData_t) + 65535, elements);
        ip_recycler = RecyclerCreate(sizeof(struct __iphdr), elements);
        
        return 0;
}
