/*****
*
* Copyright (C) 1998, 1999, 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>


#include "list.h"
#include "packet.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"
#include "detect-plugins.h"
#include "config-engine.h"
#include "pconfig.h"
#include "auth-common.h"
#include "common.h"
#include "config.h"
#include "ssl.h"
#include "daemonize.h"
#include "capture.h"
#include "compat.h"


extern Pconfig_t config;
int config_quiet;


static void print_usage(void)
{
	fprintf(stderr, "Usage :\n\n");
        fprintf(stderr, "\t-v --version Print version number.\n");
	fprintf(stderr, "\t-d --daemonize Run in daemon mode.\n");
        fprintf(stderr, "\t-P --pidfile [pidfile] Write Prelude PID to pidfile.\n");
         
        fprintf(stderr, "\t-q --quiet Quiet mode.\n");
        fprintf(stderr, "\t-u --user create new Prelude user.\n");

#ifdef HAVE_SSL
        fprintf(stderr, "\t-c --certificate create a new Prelude certificate.\n");
        fprintf(stderr, "\t-n --not-crypt Specify that the key should be stored\n"
                "\t   as is (not encrypted) on the local hardisk. This will prevent\n"
                "\t   you to be asked for a password each time you run Prelude.\n");
#endif
        fprintf(stderr, "\t-r --read-from <filename> Read file previously recorded.\n");
	fprintf(stderr, "\t-w --write-to <filename> Record packets to specified file. Don't analyze them.\n");
	fprintf(stderr, "\t-a --report-addr <server> Specify a report server.\n");
	fprintf(stderr, "\t-p --report-port <port> Specify the report server port.\n");
        fprintf(stderr, "\t-x --use-xdr Tell Prelude to use XDR to communicate with Prelude Report.\n");

        fprintf(stderr, "\t-s --snaplen specify the maximum number of bytes to capture by packet,\n"
                "the value should be set according to the interface you capture having the biggest MTU.\n");
        
        fprintf(stderr, "\t-i <interface> Specify an interface to read packet from,\n");
        fprintf(stderr, "\t\t\teventually followed by a quoted bpf rules.\n\n");

        fprintf(stderr, "Usage (plugin help):\n\n");
        fprintf(stderr, "\t-m --plugin <name> <option> to set/get plugin specific options.\n\n");

        plugin_set_args(0, NULL);
        plugins_print_opts(DETECT_PLUGIN_DIR);
}

static void configure_address(config_t *cfg) 
{
        const char *ret;
        
        if ( config.addr )
                return;
        
        ret = config_get(cfg, "Prelude", "report addr");
        if ( !ret ) 
                config.addr = "unix";
        else
                config.addr = strdup(ret);
}

static void configure_port(config_t *cfg) 
{
        const char *ret;
        
        if ( config.port != 0 )
                return;
        
        ret = config_get(cfg, "Prelude", "report port");
        if ( !ret )
                config.port = 5554;
        else
                config.port = atoi(ret);
}


static int verify_opt(config_t *cfg)
{
        if ( config.use_xdr == 1 && strcmp(config.addr, "unix") == 0) {
                fprintf(stderr,
                        "Can't use XDR when using UNIX socket: XDR aim to give the\n"
                        "possibility to communicate with a report server located on a\n"
                        "non binary compatible remote machine.\n\n");
                return -1;
        }
        
	return 0;
}


int pconfig_set(int argc, char **argv)
{
        int i = 0, ret;
        int wfile = 0, rfile = 0;
        
#ifdef HAVE_SSL
        int create_cert = 0;
#endif
        config_t *cfg;
        struct option opts[] = 
        {
                { "help", no_argument, NULL, 'h' },
                { "user", no_argument, NULL, 'u' },
                { "quiet", no_argument, NULL, 'q' },
                { "daemonize", no_argument, NULL, 'd' },
                { "pidfile", required_argument, NULL, 'P' },
                { "snaplen", required_argument, NULL, 's' },
                { "write-to", required_argument, NULL, 'o' },
                { "read-from", required_argument, NULL, 'r' },
                { "report-addr", required_argument, NULL, 'a' },
                { "report-port", required_argument, NULL, 'p' },
                { "retry-report", required_argument, NULL, 't' },
                { "interface", required_argument, NULL, 'i' },
                { "version", no_argument, NULL, 'v' },
                { "use-xdr", no_argument, NULL, 'x' },
#ifdef HAVE_SSL
                { "certificate", no_argument, NULL, 'c' },
                { "not-crypt", no_argument, NULL, 'n' },
#endif
                { 0, 0, 0, 0 }
        };

        /* Default */
	config_quiet = 0;
	config.daemonize = 0;
        config.snaplen = 8192;
	config.addr = NULL;
	config.port = 0;
        config.use_xdr = 0;
        config.ssl_key_crypt = 1;
        config.pidfile = NULL;
	/*         */
        
	while (1) {
		int c;

                c = getopt_long(argc, argv, "a:t:i:qw:r:dhus:p:m:vxcnP:", opts, NULL);
		if (c == -1)
			break;

		switch (c) {

                case 'x':
                        config.use_xdr = 1;
                        break;
                        
                case 'u':
                        /*
                         * Create a new account under authentication file 'AUTH_FILE',
                         * and not crypted (we are the client).
                         */
                        exit(auth_create_account(CONFIG_DIR"/prelude.auth", 0));

                case 'q':
			config_quiet = 1;
			break;
#ifdef HAVE_SSL
		case 'c':
                        create_cert = 1;
                        break;

                case 'n':
                        config.ssl_key_crypt = 0;
                        break;
#endif
                case 'd':
                        config_quiet = 1;
			config.daemonize = 1;
			break;

                case 'P':
                        config.pidfile = optarg;
                        break;

                case 's':
                        config.snaplen = atoi(optarg);
                        break;

                case 'w':
                        wfile = 1;
                        capture_to_file(optarg);
			break;

                case 'r':
                        ret = capture_from_file(optarg,
                                          (argv[optind] != NULL && *argv[optind] != '-')
                                          ? argv[optind] : NULL);
                        if ( ret < 0 ) {
                                log(LOG_ERR, "Couldn't select %s device for capture.\n", optarg);
                                break;
                        }

                        rfile = 1;
                        i = 1;
                        break;

                case 'i':
                        ret = capture_from_device(optarg,
                                                (argv[optind] != NULL && *argv[optind] != '-')
                                                ? argv[optind] : NULL);
                        if ( ret < 0 ) {
                                log(LOG_ERR, "Couldn't select %s device for capture.\n", optarg);
                                return -1;
                        }

                        i = 1;
			break;

                case 'a':
			config.addr = optarg;
			break;
                        
		case 'p':
			config.port = atoi(optarg);
                        break;
                        
                case 'm':
                        plugin_set_args(argc, argv);
                        goto end;
                        
		case 'h':
                        plugin_set_args(argc, argv);
                        print_usage();
			exit(0);

                case 'v':
                        printf("\n%s version %s.\n\n", PACKAGE, VERSION);
                        exit(0);
                }        
	}

  end:
        
        cfg = config_open(PRELUDE_CONF);
        if ( ! cfg ) {
                log(LOG_ERR, "opening configuration file %s.\n", PRELUDE_CONF);
                return -1;
        }

        configure_address(cfg);
        
        if ( strcmp(config.addr, "unix") != 0 ) 
                configure_port(cfg);

        
#ifdef HAVE_SSL
        if (! config.ssl_key_crypt && ! create_cert ) {
                fprintf(stderr, "-n (--not-crypt) option have to be used with the '-c' option\n");
                return -1;
        }
        
        if ( create_cert ) {
                if ( strcmp(config.addr, "unix") == 0 ) {
                        fprintf(stderr, "SSL communication is only usefull when communicating\n"
                                "to the report server throught the network. The configured\n"
                                "report server address is set to local UNIX socket.\n");
                        return -1;
                }
                exit(ssl_add_certificate());
        }        
#endif

        if ( ! i ) {
		fprintf(stderr, "\nNo interface nor file to read the packet from were specified.\n\n");
		print_usage();
		return -1;
	}

        if ( wfile && rfile ) {
		fprintf(stderr, "\nCan't both read and write packet from a file.\n\n");
                return -1;
	}

        ret = verify_opt(cfg);
        config_close(cfg);
        
	return ret;
}
