/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#include "packet.h"
#include "timer.h"
#include "common.h"
#include "alert.h"
#include "alert-prv.h"
#include "plugin-common.h"
#include "rqueue.h"
#include "recycler.h"
#include "plugin-detect.h"
#include "detect-plugins-api.h"
#include "rsend.h"


/*
 * Wait one second after last report before
 * sending it, in order to avoid duplicate on attack
 * where the IP address can't be faked (kind != normal)...
 *
 * This module is bypassed if the report 'kind'
 * is set to normal, because we don't do any check
 * for the attacking source here, and so several source would
 * be merged...
 */


/*
 * Function called back when one of the report queue
 * item expire.
 */
static void rqueue_expire(unsigned long ptr) 
{
        rqueue_t *item = (rqueue_t *) ptr;
        
        timer_destroy(&item->timer);

        log(LOG_INFO, "Debug: Flushing queued report for id=%p, count=%d...\n",
            item, alert_count(item->report));

        /*
         * This is now going to rsend, which will itself
         * be responssible for calling the asynchronous machinery.
         *
         * item->report and packet_release will be called
         * in the async IO thread.
         */
        rsend_emmit(item->report);
        item->report = NULL;
}



/*
 * First time we add this to our report queue.
 */
static void rqueue_first(rqueue_t *item) 
{
        alert_count(item->report) = 1;
        timer_set_expire(&item->timer, 1);
        timer_set_callback(&item->timer, rqueue_expire);
        timer_set_data(&item->timer, item);       
        timer_init(&item->timer);
}



/*
 * Update one of our report queue item.
 */
__inline__ static void rqueue_update(rqueue_t *item) 
{
        alert_time_end(item->report) = time(NULL);
        alert_count(item->report)++;
        timer_reset(&item->timer);
}



/*
 *
 */
static void rqueue_generic(rqueue_t *item) 
{
        /*
         * As the packet will be passed to the async IO subsystem,
         * it should be locked until it is written to Prelude Report.
         */
        packet_lock(item->report->packet);
        
        /*
         * if kind is normal, source IP can't be faked,
         * so we must do one report for one attack...
         * do not queue it.
         */
        if ( item->report->kind == normal ) {
                alert_count(item->report) = 1;
                log(LOG_INFO, "Debug: Flushing report for id=%p, count=%d...\n",
                    item->report, alert_count(item->report));
                rsend_emmit(item->report);
                item->report = NULL;
        } else 
                rqueue_first(item);
}


static int report_new(rqueue_t *item, Packet_t *packet,
                      rkind_t kind, char *quickmsg, size_t qmsglen) 
{
        alert_t *report;
        
        if ( item->report && kind != normal ) {
                rqueue_update(item);
                return -1;
        } else {
                report = item->report = malloc(sizeof(alert_t));
                if (! item->report ) {
                        log(LOG_ERR, "couldn't allocate memory.\n");
                        return -1;
                }
                
                alert_quickmsg(report) = quickmsg;
                alert_quickmsg_len(report) = qmsglen;
                                
                alert_time_start(report) = time(NULL);
                alert_time_end(report) = 0;
                alert_packet(report) = packet;
                alert_kind(report) = kind;
        }

        return 0;
}



/*
 * Queue a report depending on it's kind.
 */
void plugin_rqueue_report(rqueue_t *item, plugin_generic_t *plugin,
                          Packet_t *packet, rkind_t kind, char *qmsg,
                          size_t qmsglen, const char *msg, ...) 
{
        int ret;
        va_list va;
        
        ret = report_new(item, packet, kind, qmsg, qmsglen);
        if ( ret < 0 )
                return;
        
        va_start(va, msg);
        ret = vsnprintf(alert_message(item->report),
                        sizeof(alert_message(item->report)), msg, va);
        va_end(va);

        /*
         * return -1 if the output was truncated due to this limit. (Thus until
         * glibc 2.0.6. Since glibc 2.1 these functions follow the
         * C99 standard and return the number of characters (excluding
         * the trailing '\0') which would have been written to
         * the final string if enough space had been available.)
         */
        if ( ret > sizeof(alert_message(item->report)) || ret < 0 )
                alert_message_len(item->report) = sizeof(alert_message(item->report));
        else
                alert_message_len(item->report) = ret + 1;
        
        alert_from(item->report) = plugin_name(plugin);
        alert_from_len(item->report) = plugin_name_len(plugin);

        alert_desc(item->report) = plugin_desc(plugin);
        alert_desc_len(item->report) = plugin_desc_len(plugin);
        
        alert_author(item->report) = plugin_author(plugin);
        alert_author_len(item->report) = plugin_author_len(plugin);

        alert_contact(item->report) = plugin_contact(plugin);
        alert_contact_len(item->report) = plugin_contact_len(plugin);
        
        rqueue_generic(item);
}



/*
 *
 */
void prelude_rqueue_report(rqueue_t *item, Packet_t *packet,
                           rkind_t kind, char *qmsg,
                           size_t qmsglen, const char *msg, ...) 
{
        int ret;
        va_list va;

        ret = report_new(item, packet, kind, qmsg, qmsglen);
        if ( ret < 0 )
                return;

        va_start(va, msg);
        ret = vsnprintf(alert_message(item->report),
                  sizeof(alert_message(item->report)), msg, va);
        va_end(va);

        /*
         * return -1 if the output was truncated due to this limit. (Thus until
         * glibc 2.0.6. Since glibc 2.1 these functions follow the
         * C99 standard and return the number of characters (excluding
         * the trailing '\0') which would have been written to
         * the final string if enough space had been available.)
         */
        if ( ret > sizeof(alert_message(item->report)) || ret < 0 )
                alert_message_len(item->report) = sizeof(alert_message(item->report));
        else
                alert_message_len(item->report) = ret + 1;
        
        alert_from(item->report) = "Prelude Core";
        alert_from_len(item->report) = 13;

        alert_desc(item->report) = "Prelude Core detection engine";
        alert_desc_len(item->report) = 30;
        
        alert_author(item->report) = "The Prelude Team";
        alert_author_len(item->report) = 17;

        alert_contact(item->report) = "prelude-devel@mandrakesoft.com";
        alert_contact_len(item->report) = 31;
        
        rqueue_generic(item);
}











