/*****
*
* Copyright (C) 1999 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <signal.h>
#include <sys/uio.h>
#include <assert.h>
#include <sys/ioctl.h>

#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "pconfig.h"
#include "rsend.h"
#include "rsave.h"
#include "common.h"
#include "timer.h"
#include "auth.h"
#include "async-write.h"
#include "detect-plugins-api.h"
#include "plugin-common.h"
#include "plugin-common-prv.h"
#include "rwrite.h"
#include "capture.h"
#include "socket-op.h"

#ifdef HAVE_SSL
#include "ssl.h"
#endif

#define UNIX_SOCK "/var/lib/prelude/socket"

#define RECONNECT_TIME_WAIT 10
#define RECONNECT_TIME_EXPAND(current_timeout) (current_timeout) * 2


extern Pconfig_t config;
static int sock;
static void report_timer_set(void);


static void sigpipe_handler(int signo) 
{
        log(LOG_ERR, "PID %d caught pipe signal.\n", getpid());
        close(sock);
        sock = -1;
        rwrite_exit();
        report_timer_set();
}



/*
 * Connect to the specified address in a generic manner
 * (can be Unix or Inet ).
 */
static int generic_connect(struct sockaddr *addr, size_t addr_size) 
{
        int ret;
        
        ret = fcntl(sock, F_SETOWN, getpid());
        if ( ret < 0 ) {
                log(LOG_ERR, "couldn't set children to receive signal.\n");
                return -1;
        }

        ret = connect(sock, addr, addr_size);
	if ( ret < 0 ) {
                log(LOG_ERR,"error connecting socket.\n");
                return -1;
	}
        
	return 0;
}



/*
 * Setup an Unix connection.
 */
static int unix_connect(void)
{
        int ret;
    	struct sockaddr_un addr;
        
	log(LOG_INFO, "\t- Connecting to Unix prelude report server.\n");
        
	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if ( sock < 0 ) {
                log(LOG_ERR,"Error opening unix socket.\n");
		return -1;
	}
        
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, UNIX_SOCK);

        ret = generic_connect((struct sockaddr *)&addr, sizeof(addr));
        if ( ret < 0 ) {
                log(LOG_ERR,"Error connecting to report server.\n");
                close(sock);
                return -1;
        }

	return 0;
}




static int ignore_report_server_connection(struct sockaddr_in *saddr,
                                           struct sockaddr_in *daddr) 
{
        int r;
        char buf[1024];

        r = snprintf(buf, sizeof(buf), "(!(host %s and ", inet_ntoa(saddr->sin_addr));
        snprintf(&buf[r], sizeof(buf) - r, "host %s and port %u and port %u))",
                 inet_ntoa(daddr->sin_addr),
                 ntohs(saddr->sin_port), ntohs(daddr->sin_port));
        
        r = capture_setup_global_bpf(buf);
        if ( r < 0 ) {
                log(LOG_ERR, "Couldn't create a BPF rule to ignore report server connection.\n");
                return -1;
        }

        return 0;
}




/*
 * Setup an Inet connection.
 */
static int inet_connect(const char *server, unsigned int port)
{
        int ret, on, len;
	struct sockaddr_in daddr, saddr;
        
	log(LOG_INFO, "\t- Connecting to Tcp prelude report server.\n");

        sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if ( sock < 0 ) {
		log(LOG_ERR,"Error opening inet socket.\n");
		return -1;
	}

	daddr.sin_family = AF_INET;
	daddr.sin_port = htons(port);
	daddr.sin_addr.s_addr = inet_addr(server);

        /*
         * We want packet to be sent as soon as possible,
         * this mean not using the Nagle algorithm which try to minimize
         * packets on the network at the expense of buffering the data.
         */
        on = 1;
        ret = setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(int));
        if ( ret < 0 ) 
                log(LOG_ERR, "couldn't turn the tcp Nagle algorithm off.\n");
        
        
        ret = generic_connect((struct sockaddr *)&daddr, sizeof(daddr));
        if ( ret < 0 ) {
                log(LOG_ERR,"Error connecting to %s.\n", server);
                goto err;
        }

        len = sizeof(saddr);
        ret = getsockname(sock, (struct sockaddr *)&saddr, &len);
        if ( ret < 0 ) {
                log(LOG_ERR, "couldn't get source connection informations.\n");
                goto err;
        }
        
        ret = ignore_report_server_connection(&saddr, &daddr);
        if ( ret < 0 ) {
                log(LOG_ERR, "Couldn't create a BPF rule to ignore report server connection.\n");
                goto err;
        }

        return 0;

 err:
        close(sock);
        return -1;
}




static int set_options(const char *optbuf) 
{
        config.use_ssl = 0;
        
#ifdef HAVE_SSL
        if ( strstr(optbuf, "ssl=supported;") ) {
                log(LOG_INFO, "\t- Report Server support SSL, initializing SSL subsystem.\n");
                config.use_ssl = 1;
        } else
                log(LOG_INFO, "\t- Report Server do not support SSL, not using encryption.\n");
#endif


#if HAVE_XDR
        if ( config.use_xdr ) {
                if ( strstr(optbuf, "xdr=supported;") ) 
                        log(LOG_INFO, "Report Server support XDR, enabling it as requested.\n");
                else {
                        log(LOG_INFO, "Error : Report Server doesn't support XDR but you requested it.\n");
                        return -1;
                }
        }
#endif
        
        return 0;
}




/*
 * Report server should send us a string looking like :
 * xdr=xxx; ssl=xxx;
 *
 * with xxx having the value "supported" or "unsupported"
 *
 * The string have to be terminated with a '\n\0'.
 */
static int setup_connection(void) 
{
        char buf[1024];
        int len = 0, ret;
        const char *ssl, *xdr;
        
        ret = socket_read_delimited(sock, buf, sizeof(buf));
        if ( ret < 0 ) {
                log(LOG_ERR, "error waiting for Report server config string.\n");
                return -1;
        }
        
        ret = set_options(buf);
        if ( ret < 0 )
                return -1;
        
        ssl = (config.use_ssl) ? "yes" : "no";
        xdr = (config.use_xdr) ? "yes" : "no";
        len = snprintf(buf, sizeof(buf), "use_ssl=%s; use_xdr=%s;\n", ssl, xdr);
      
	ret = socket_write_delimited(sock, buf, ++len);
	if ( ret != len ) {
                log(LOG_ERR, "couldn't write config string to Report Server.\n");
                return -1;
        }
        
        return 0;
}




static int do_connect(const char *server, int port) 
{
        int ret;
        
        ret = strcmp(server, "unix");
	if (ret == 0) 
		return unix_connect();
	
        ret = inet_connect(server, port);
        if ( ret < 0 )
                return -1;

        ret = setup_connection();
        if ( ret < 0 ) 
                goto err;

#ifdef HAVE_SSL
        if ( config.use_ssl == 1 ) {
                static int ssl_initialized = 0;
                
                if ( ! ssl_initialized ) {
                        ret = ssl_init_client();
                        if ( ret < 0 )
                                goto err;
                        
                        ssl_initialized = 1;
                }
                
                ret = ssl_connect_server(sock);
                if ( ret < 0 )
                        goto err;
        } else 
#endif
                if ( auth_init() < 0 || auth_client(sock) < 0 )
                        goto err;
        
        return 0;

 err:
        close(sock);
        return -1;
        
}



/*
 * function called back when a report timer expire.
 */
static void expire(unsigned long data) 
{
        int ret;
        timer_t *timer = (timer_t *) data;
        
        ret = do_connect(config.addr, config.port);
        if ( ret < 0 ) {
                log(LOG_ERR,"can't reconnect to report server.\n");
                timer_set_expire(timer, RECONNECT_TIME_EXPAND(timer_expire(timer)));
                timer_reset(timer);
                return;
        }
        
        timer_destroy(timer);

        /*
         * Re initialize rwrite,
         * if the way data is written (xdr for exemple) has changed.
         * This could be the case if, for exemple, the administrator changed
         * the kind of machine that host Prelude Report.
         */
        rwrite_init(sock, getpid());
        backout_report(sock);
}




static void report_timer_set(void) 
{
        static timer_t timer;
        
        timer_set_callback(&timer, expire);
        timer_set_data(&timer, &timer);
        timer_set_expire(&timer, RECONNECT_TIME_WAIT);

        timer_init(&timer);
}



/**
 * rsend_emmit:
 * @report: A report to emmit.
 *
 * FIXME: In future version, caller should use async_write directly. 
 */
void rsend_emmit(alert_t *report)
{
        struct timeval b, e;
        static double total = 0;
        static int i = 0;
        
        gettimeofday(&b, NULL);
        
        async_write(report);
                
        gettimeofday(&e, NULL);

        i++;
        total += (e.tv_sec + (double) (e.tv_usec * 1e-6)) -
            (b.tv_sec  + (double) (b.tv_usec * 1e-6));
        
        if ( i == 100 ) {
                log(LOG_INFO, "Debug: %fs a report, (avr on %d report)\n", total / i, i);
                total = i = 0;
        }
}



/**
 * rsend_init:
 *
 * Init the 'report send' module,
 * This connect to the report server, negociate XDR and SSL
 * communication.
 *
 * Returns: 0 on success, -1 otherwise. 
 */
int rsend_init(void)
{
        int ret;

        signal(SIGPIPE, sigpipe_handler);
        
        ret = do_connect(config.addr, config.port);
        if ( ret < 0 )
                return -1;

        rwrite_init(sock, getpid());
        
        ret = async_init();
        if ( ret < 0 ) {
                log(LOG_ERR, "coudln't initialize Asynchronous IO subsystem.\n");
                return -1;
        }
        
	return 0;
}


