/*****
*
* Copyright (C) 1998 - 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>

#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "pconfig.h"
#include "detect-plugins-api.h"
#include "common.h"
#include "rsave.h"
#include "socket-op.h"
#include "rwrite.h"
#include "write-func.h"


#ifdef HAVE_SSL
  #include "ssl.h"
#endif

#ifdef HAVE_XDR
  #include <rpc/types.h>
  #include <rpc/xdr.h>
  #include "rxdr.h"
  static XDR xdrs;
#endif


/* FD to write data to. */
static int fd;

/* PID of the process that should recv error notification. */
static pid_t ppid;

/* ptr on an alert writing function. */ 
static int (*write_func_ptr)(int fd, alert_t *alert);

/* ptr on a real write function */
static ssize_t (*my_write)(int fd, const void *buf, size_t size);


extern Pconfig_t config;



/*
 * Wrapper function to use for XDR decoding.
 */
#ifdef HAVE_XDR

static int rxdr_encode_wrapper(int fd, alert_t *alert) 
{
        return rxdr_encode(alert, &xdrs);
}



static int write_xdr_report(char *handle, char *buf, int len) 
{
        return my_write(fd, buf, len);
}

#endif



/**
 * rwrite_write:
 * @alert: The report to send.
 * @backup_on_error: A flag specifying if data should be backuped on error.
 *
 * This function try to write a report to the file descriptor
 * passed to the rwrite_init() function.
 *
 * In case of error, the process specified in the rwrite_init() call,
 * is notified by a SIGPIPE, telling it that the connection with the
 * Report Server was eventually closed.
 *
 * In case rwrite_exit() was called or if an error occur,
 * and if the backup_on_error flag is set to TRUE, the report is backuped.
 *
 * Returns: 0 on success, -1 otherwise.
 */
int rwrite_write(alert_t *alert, int backup_on_error) 
{
        int ret = 0;

        if ( fd == -1 && backup_on_error )
                ret = backup_report(alert);
        else {
                ret = write_func_ptr(fd, alert);
                if ( ret < 0 ) {
                        log(LOG_ERR, "error trying to send the report.\n");
                        kill(ppid, SIGPIPE);
                        if ( backup_on_error )
                                backup_report(alert);
                }
        }
        
        return ret;
}




/**
 * rwrite_init:
 * @_fd: A file descriptor to be used.
 * @_ppid: The PID of the process where notification should be sent.
 *
 * Initialize the rwrite module.
 *
 */
void rwrite_init(int _fd, pid_t _ppid) 
{
        fd = _fd;
	ppid = _ppid;
            
        my_write = write;
        
#ifdef HAVE_SSL
        if ( config.use_ssl == 1 )
                my_write = ssl_write;
#endif

#ifdef HAVE_XDR
        if ( config.use_xdr == 1 ) {
                xdrrec_create(&xdrs, 0, 0, (char *)&fd, NULL, write_xdr_report);
                xdrs.x_op = XDR_ENCODE;
                write_func_ptr = rxdr_encode_wrapper;
        } else
#endif
            
#ifdef HAVE_SSL
                write_func_ptr = (config.use_ssl == 1 ) ? write_raw_report : writev_raw_report;
#else
                write_func_ptr = writev_raw_report;
#endif
}



/**
 * rwrite_exit:
 *
 * Uninitialize the rwrite module.
 * Further call to the rwrite_write() function
 * will result the data being backuped.
 *
 */
void rwrite_exit(void) 
{
        fd = -1;
        
#ifdef HAVE_XDR
        if ( config.use_xdr == 1 )
                xdr_destroy(&xdrs);
#endif
}







